@@ -945,6 +945,9 @@ int add_mbox_messages(struct database *db)/*{{{*/
             printf("Scanning %s[%d] at [%d,%d)\n", mb->path, j, (int)start, (int)(start + len));
           }
           db->msgs[n].date = r8->hdrs.date;
+          db->msgs[n].seen = r8->hdrs.flags.seen;
+          db->msgs[n].replied = r8->hdrs.flags.replied;
+          db->msgs[n].flagged = r8->hdrs.flags.flagged;
           tokenise_message(n, db, r8);
           free_rfc822(r8);
         } else {
